# file generated by setuptools_scm
# don't change, don't track in version control
__version__ = version = '0.1.dev600+g6605467.d20230217'
__version_tuple__ = version_tuple = (0, 1, 'dev600', 'g6605467.d20230217')
